'*****************************************************************************
'*  Name    : I2C Library                                                    *
'*  Author  : David John Barker                                              *
'*  Notice  : Copyright (c) 2006 Mecanique                                   *
'*          : All Rights Reserved                                            *
'*  Date    : 05/04/06                                                       *
'*  Version : 1.1                                                            *
'*  Notes   : 1.1 Added 'WaitForWrite()' routine                             *
'*          : 1.0 Initial release                                            *
'*****************************************************************************
Module I2C

Include "system.bas"

// map registers to SSP(x)
#if _mssp = 0
   #error _device + " does not support MSSP"

// single SSP...
#elseif _mssp = 1
Dim
   SSPControl1 As SSPCON1,
   SSPControl2 As SSPCON2,
   SSPStatus As SSPSTAT,
   SSPBuffer As SSPBUF,
   SSPAddress As SSPADD,
   SCL As PORTC.3,
   SDA As PORTC.4  
   
// has more than one SSP module...   
#else
Dim                           // -> MSSP2
   SSPControl1 As SSP1CON1,   //    As SSP2CON1
   SSPControl2 As SSP1CON2,   //    As SSP2CON2
   SSPStatus As SSP1STAT,     //    As SSP2STAT
   SSPBuffer As SSP1BUF,      //    As SSP2BUF
   SSPAddress As SSP1ADD,     //    As SSP2ADD
   SCL As PORTC.3,            //    As PORTD.6
   SDA As PORTC.4             //    As PORTD.5
#endif

// SSPSTAT bitnames, master mode only...
Public Dim
   SMP As SSPStatus.7,     // slew rate control bit
   CKE As SSPStatus.6,     // SMBus select
   P As SSPStatus.4,       // stop bit
   S As SSPStatus.3,       // start bit
   RW As SSPStatus.2,      // transmit in progress
   BF As SSPStatus.0       // buffer full (receive and transmit)    

// SSPCON1 bitnames, master mode only...
Public Dim
   WCOL As SSPControl1.7,  // collision detect
   SSPOV As SSPControl1.6, // receive overflow
   SSPEN As SSPControl1.5, // synchronous receive enable
   
   // synchronous mode select bits, %1000 for master mode
   SSPM3 As SSPControl1.3, 
   SSPM2 As SSPControl1.2, 
   SSPM1 As SSPControl1.1, 
   SSPM0 As SSPControl1.0 

// SSPCON2 bitnames, master mode only...
Public Dim
   ACKSTAT As SSPControl2.6, // acknowledge status bit
   ACKDT As SSPControl2.5,   // acknowledge data bit
   ACKEN As SSPControl2.4,   // acknowledge sequence enable bit
   RCEN As SSPControl2.3,    // receive enable bit
   PEN As SSPControl2.2,     // stop condition enable bit
   RSEN As SSPControl2.1,    // repeated start condition enabled bit
   SEN As SSPControl2.0      // start condition enabled  
   
// I2C constants...
const FOSC = _clock * 1000000
Public const 
   I2C_SLEW_OFF = $C0,                      // slew rate disabled for 100kHz mode 
   I2C_SLEW_ON = $00,                       // slew rate enabled for 400kHz mode
   I2C_100_KHZ = FOSC / (100000 * 4) - 1,   // 100 KHz
   I2C_400_KHZ = FOSC / (400000 * 4) - 1,   // 400 KHz
   I2C_1000_KHZ = FOSC / (1000000 * 4) - 1, // 1 MHz
   I2C_ACKNOWLEDGE = 0,                     // acknowledge data bit (acknowledge)
   I2C_NOT_ACKNOWLEDGE = 1                  // acknowledge data bit (NOT acknowledge)

// local helper aliases...   
Dim
   BufferIsFull As SSPStatus.Booleans(0),       // BF As boolean
   AckEnabled As SSPControl2.Booleans(4),       // ACKEN As boolean
   ReceiveEnabled As SSPControl2.Booleans(3),   // RCEN As boolean
   StopEnabled As SSPControl2.Booleans(2),      // PEN As boolean
   RestartEnabled As SSPControl2.Booleans(1),   // RSEN As boolean
   StartEnabled As SSPControl2.Booleans(0)      // SEN As boolean

// Public boolean alias...
Public Dim
   NotAcknowledged As SSPControl2.Booleans(6)  // ACKSTAT As boolean
{
****************************************************************************
* Name    : Initialize                                                     *
* Purpose : Initialize SSP module for I2C bus                              *
****************************************************************************
}         
Public Sub Initialize(pBaudrate As byte = I2C_100_KHZ, pSlew As byte = I2C_SLEW_OFF)
  SSPAddress = pBaudrate     // set baudrate
  SSPStatus = pSlew          // POR state, optional slew
  SSPControl2 = $00          // POR state
  input(SCL)                 // set SCL (clock pin) to input
  input(SDA)                 // set SDA (data pin) to input
  SSPControl1 = $28          // master mode, enable synchronous serial port
End Sub
{
****************************************************************************
* Name    : IsIdle                                                         *
* Purpose : Returns the current I2C idle state                             *
****************************************************************************
}         
Public function IsIdle() As boolean
   Result = (SSPControl2 and $1F) = 0 and (RW = 0)
end function
{
****************************************************************************
* Name    : WaitForIdle                                                    *
* Purpose : Wait until the I2C bus is in an idle state                     *
****************************************************************************
}         
Public sub WaitForIdle()
   repeat
      ClrWDT
   until IsIdle
End Sub
{
****************************************************************************
* Name    : Start                                                          *
* Purpose : Send an I2C bus start condition                                *
****************************************************************************
}         
Public Sub Start()
   WaitForIdle
   StartEnabled = True
   While StartEnabled
   Wend
End Sub
{
****************************************************************************
* Name    : Stop                                                           *
* Purpose : Send an I2C bus stop condition                                 *
****************************************************************************
}         
Public Sub Stop()
  WaitForIdle
  StopEnabled = true
  while StopEnabled
  wend
End Sub
{
****************************************************************************
* Name    : Restart                                                        *
* Purpose : Send an I2C bus restart condition                              *
****************************************************************************
}         
Public Sub Restart()
   WaitForIdle
   RestartEnabled = true
   while RestartEnabled
   wend
End Sub
{
****************************************************************************
* Name    : Acknowledge                                                    *
* Purpose : Initiate I2C acknowledge                                       *
*         : pAck = 1, NOT acknowledge                                      *
*         : pAck = 0, acknowledge                                          *
****************************************************************************
}         
Public Sub Acknowledge(pAck As bit = I2C_ACKNOWLEDGE)
  WaitForIdle
  ACKDT = pAck
  AckEnabled = true
  while AckEnabled
  wend
End Sub
{
****************************************************************************
* Name    : ReadByte (OVERLOAD)                                            *
* Purpose : Read a single byte from the I2C bus                            *
****************************************************************************
}         
Public function ReadByte() As byte
  WaitForIdle
  ReceiveEnabled = true
  Repeat
  Until BufferIsFull
  Result = SSPBuffer
end function
{
****************************************************************************
* Name    : ReadByte (OVERLOAD)                                            *
* Purpose : Read a single byte from the I2C bus, with Acknowledge          *
****************************************************************************
}         
Public function ReadByte(pAck As bit) As byte
  WaitForIdle
  ReceiveEnabled = true
  Repeat
  Until BufferIsFull
  Result = SSPBuffer
  Acknowledge(pAck)
end function
{
****************************************************************************
* Name    : WriteByte                                                      *
* Purpose : Write a single byte to the I2C bus                             *
****************************************************************************
}         
Public Sub WriteByte(pData As Byte)
   WaitForIdle
   SSPBuffer = pData
   While BufferIsFull
   Wend
End Sub
{
****************************************************************************
* Name    : WaitForWrite                                                   *
* Purpose : Wait until write sequence has completed                        *
****************************************************************************
}         
Public Sub WaitForWrite(pControl As byte)
   Dim Timeout As byte
   Timeout = $FF
   
   Start
   WriteByte(pControl)
   WaitForIdle
   while NotAcknowledged and (Timeout > 0)
      Restart
      WriteByte(pControl)
      WaitForIdle
      dec(Timeout)
   wend
   Stop
End Sub

